package de.tudarmstadt.maki.simonstrator.api.component.vehicular.data.map;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import de.tudarmstadt.maki.simonstrator.api.Event;
import de.tudarmstadt.maki.simonstrator.api.EventHandler;
import de.tudarmstadt.maki.simonstrator.api.Randoms;
import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.location.Location;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.data.change.ChangeContainer;

public class ComplexSensorValueMap implements SensorValueMap, EventHandler {

	private static final double SCALE_FACTOR = 10;

	private Random _random = Randoms.getRandom(getClass());

	private final long INTERVAL = Time.MINUTE;
	private long _lastTrendSet = 0;
	private double _trend = 0;

	private double[][] _mean;
	private double _max;

	private Map<Long, MeanValueContainer> _pastValues = new HashMap<>();
	private List<Long> _availableTimestamps = new ArrayList<>();

	private double _temporalChange;
	private double _geographicalChange;

	private final double _worldWidth;
	private final double _worldHeight;

	public ComplexSensorValueMap(double pMean, double pMax, double pTemporalChange, double pGeographicalChange,
			double pWorldWidth, double pWorldHeight) {

		_worldWidth = pWorldWidth;
		_worldHeight = pWorldHeight;

		_mean = new double[(int) (_worldWidth / STEP_SIZE + 1)][(int) (_worldHeight / STEP_SIZE + 1)];

		for (int i = 0; i < _mean.length; i++) {
			Arrays.fill(_mean[i], pMean);
		}

		updatePastValuesMap();

		setTrend();

		_max = pMax;

		_temporalChange = pTemporalChange / SCALE_FACTOR;
		_geographicalChange = pGeographicalChange / SCALE_FACTOR;

		Event.scheduleWithDelay(Time.SECOND, this, null, 0);
	}

	private void setTrend() {
		_trend = _random.nextDouble();
		_lastTrendSet = Time.getCurrentTime();
	}

	@Override
	public double getValueAt(Location pLocation) {
		return getValueAt(pLocation, _mean);
	}

	@Override
	public double getValueAt(Location pLocation, long pTimestamp) {
		for (int i = _availableTimestamps.size() - 1; i >= 0; i--) {
			long timestamp = _availableTimestamps.get(i);
			if (timestamp <= pTimestamp) {
				MeanValueContainer oldMean = _pastValues.get(timestamp);

				return getValueAt(pLocation, oldMean.getMean());
			}
		}
		return Double.NEGATIVE_INFINITY;
	}

	private double getValueAt(Location pLocation, double[][] pMean) {
		double x = pLocation.getLongitude();
		double y = pLocation.getLatitude();

		int indexX = (int) (x / STEP_SIZE);
		int indexY = (int) (y / STEP_SIZE);

		return pMean[indexX][indexY];
	}

	@Override
	public void eventOccurred(Object pContent, int pType) {
		if (Time.getCurrentTime() - INTERVAL >= _lastTrendSet) {
			setTrend();
		}

		for (int x = 0; x < _mean.length; x++) {
			for (int y = 0; y < _mean[x].length; y++) {
				performTimelyChanged(x, y);
			}
		}

		updatePastValuesMap();

		Event.scheduleWithDelay(Time.SECOND, this, null, 0);
	}

	private void performTimelyChanged(int pX, int pY) {
		double oldMean = _mean[pX][pY];
		double nextDouble = _random.nextDouble();

		if (pX != 0 || pY != 0) {
			double maxAvailableValue = oldMean + _temporalChange / 2.0;
			double minAvailableValue = oldMean - _temporalChange / 2.0;

			if (pX > 0) {
				double adjancend = _mean[pX - 1][pY];

				double maxGeoDistance = STEP_SIZE / (double) ChangeContainer.LOCATION_INTERVAL * _geographicalChange;

				double maxAvailableValueGeo = adjancend + maxGeoDistance;
				double minAvailableValueGeo = adjancend - maxGeoDistance;

				if (maxAvailableValue > maxAvailableValueGeo) {
					maxAvailableValue = maxAvailableValueGeo;
				}
				if (minAvailableValue < minAvailableValueGeo) {
					minAvailableValue = minAvailableValueGeo;
				}
			}
			if (pY > 0) {
				double adjancend = _mean[pX][pY - 1];

				double maxGeoDistance = STEP_SIZE / (double) ChangeContainer.LOCATION_INTERVAL * _geographicalChange;

				double maxAvailableValueGeo = adjancend + maxGeoDistance;
				double minAvailableValueGeo = adjancend - maxGeoDistance;

				if (maxAvailableValue > maxAvailableValueGeo) {
					maxAvailableValue = maxAvailableValueGeo;
				}
				if (minAvailableValue < minAvailableValueGeo) {
					minAvailableValue = minAvailableValueGeo;
				}
			}
			if (pX > 0 && pY > 0) {

			}

			double newValue = Math
					.max(Math.min((maxAvailableValue - minAvailableValue) * nextDouble + minAvailableValue, _max), 0);

			_mean[pX][pY] = newValue;
		} else {
			double change;
			double factor;
			if (nextDouble < _trend) {
				factor = nextDouble / _trend;
			} else {
				factor = (1 - nextDouble) / (1 - _trend);
			}
			if (nextDouble < _trend) {
				change = factor * (-_temporalChange / 2.0);
			} else {
				change = factor * (_temporalChange / 2.0);
			}

			double newValue = Math.max(Math.min(Math.max(_mean[pX][pY] + change, 0), _max), 0);

			_mean[pX][pY] = newValue;
		}
	}

	public void updatePastValuesMap() {
		_pastValues.put(Time.getCurrentTime(), new MeanValueContainer(_mean));
		_availableTimestamps = new ArrayList<>(_pastValues.keySet());
		Collections.sort(_availableTimestamps);
	}

	@Override
	public double getMaxValue() {
		return _max;
	}

	private class MeanValueContainer {

		private double[][] _storedMean;

		private MeanValueContainer(double[][] pMean) {
			_storedMean = pMean;
		}

		private double[][] getMean() {
			return _storedMean;
		}
	}

}
